"""Compatibility layer for PyPy3.5

"""

import platform

import builtins

import runpy

import shutil

f"In >=2018, you should use a Python supporting f-strings!"

implementation = platform.python_implementation()

if implementation == "PyPy":

    def open(file, *args, **kwargs):
        return builtins.open(str(file), *args, **kwargs)

    def run_path(path):
        return runpy.run_path(str(path))

    def rmtree(path):
        shutil.rmtree(str(path))

    def fspath(path):
        return str(path)


else:

    from builtins import open
    from runpy import run_path
    from shutil import rmtree
    from os import fspath


__all__ = ["open", "run_path", "rmtree", "fspath"]



"""Command line transonic
============================

Internal API
------------

.. autofunction:: run

.. autofunction:: parse_args

"""

import argparse
from pathlib import Path
from glob import glob
import sys

from . import __version__
from .transpiler import make_backend_files
from .log import logger
from .pythranizer import compile_extensions, ext_suffix
from .util import has_to_build, clear_cached_extensions
from .compat import fspath

try:
    import pythran
except ImportError:
    pythran = False


doc = """
transonic: easily speedup your Python code with Pythran

"""


def run():
    """Run the transonic commandline

    See :code:`transonic -h`
    """
    args = parse_args()

    if args.version:
        print(__version__)
        return

    if not args.path and not args.clear_cache:
        logger.warning("No python files given. Nothing to do! ✨ 🍰 ✨.")
        return

    if args.clear_cache:
        clear_cached_extensions(args.clear_cache, force=args.force)
        return

    if args.verbose is None:
        logger.set_level(None)
    elif args.verbose == 1:
        logger.set_level("info")
    elif args.verbose > 1:
        logger.set_level("debug")
    logger.info(args)

    path = args.path

    if isinstance(path, list) and len(path) == 1:
        path = path[0]

    if isinstance(path, list):
        paths = path
    else:
        path = Path(path)
        if path.is_file():
            paths = (path,)
        elif path.is_dir():
            paths = path.glob("*.py")
        else:
            paths = glob(fspath(path))

    if not paths:
        logger.error(f"No input file found (args.path = {args.path})")
        sys.exit(1)

    make_backend_files(paths, force=args.force)

    if not pythran or args.no_pythran:
        return

    # find pythran files not already compiled
    pythran_paths = []
    for path in paths:
        path = Path(path)
        pythran_path = path.parent / "__pythran__" / path.name
        ext_path = pythran_path.with_suffix(ext_suffix)
        if has_to_build(ext_path, pythran_path):
            pythran_paths.append(pythran_path)

    compile_extensions(
        pythran_paths, args.pythran_flags, parallel=True, force=args.force
    )


def parse_args():
    """Parse the arguments"""
    parser = argparse.ArgumentParser(
        description=doc, formatter_class=argparse.RawDescriptionHelpFormatter
    )
    parser.add_argument("path", help="Path file or directory.", nargs="*")

    parser.add_argument(
        "-f",
        "--force",
        help="write the file even if it is up-to-date",
        action="store_true",
    )

    parser.add_argument(
        "-V", "--version", help="print version and exit", action="store_true"
    )

    parser.add_argument("-v", "--verbose", help="verbose mode", action="count")

    parser.add_argument(
        "-np",
        "--no-pythran",
        help="do not compile the Pythran files with Pythran",
        action="store_true",
    )

    parser.add_argument(
        "-pf",
        "--pythran-flags",
        help=(
            "Flags sent to Pythran. "
            'Default is "-march=native -DUSE_XSIMD". '
            "There has to be atleast one space in the passed string! "
            'Example: transonic toto.py -pf "-march=native "\n'
        ),
        type=str,
        default="",
    )

    parser.add_argument(
        "-cc",
        "--clear-cache",
        help=("Clear the cached extensions for a module"),
        type=str,
        # default="",
    )

    return parser.parse_args()


if __name__ == "__main__":
    run()



# cython: language_level=3
import numpy as np
cimport numpy as np

DTYPE = np.float64
ctypedef np.float64_t DTYPE_t


cpdef np.ndarray[DTYPE_t, ndim=2] add( DTYPE_t[:, ::1] f, DTYPE_t[:, ::1] g):
    cdef int i, j
    cdef int imax = f.shape[0]
    cdef int jmax = f.shape[1]
    cdef np.ndarray[DTYPE_t, ndim=2] r = np.zeros([imax, jmax], dtype=DTYPE)
    for i in range(imax):
        for j in range(jmax):
            r[i, j] = f[i, j] + g[i, j]

    return r



"""Make the Pythran files for ahead-of-time compilation
=======================================================

User API
--------

.. autofunction:: make_backend_files

Internal API
------------

.. autofunction:: parse_py_code

.. autofunction:: get_code_functions

.. autofunction:: get_codes_from_functions

.. autofunction:: make_pythran_code_functions

.. autofunction:: make_pythran_code

.. autofunction:: make_pythran_file

.. autofunction:: mock_modules

"""

from tokenize import (
    tokenize,
    untokenize,
    COMMENT,
    INDENT,
    DEDENT,
    STRING,
    NAME,
    OP,
)

from token import tok_name
import inspect
from io import BytesIO
from runpy import run_module, run_path
import sys
from unittest.mock import MagicMock as Mock
from contextlib import contextmanager
from pathlib import Path

from typing import Iterable, Optional

try:
    import black
except ImportError:
    black = False

from .log import logger
from .annotation import compute_pythran_types_from_valued_types
from .util import (
    has_to_build,
    get_source_without_decorator,
    find_module_name_from_path,
)
from .compat import open, fspath
import transonic


def parse_py_code(code: str):
    """Parse the code of a .py file and return data"""

    blocks = []
    signatures_blocks = {}
    code_blocks = {}

    functions = set()
    signatures_func = {}

    imports = []

    has_to_find_name_block = False
    has_to_find_signatures = False
    has_to_find_code_block = False
    in_def = False

    g = tokenize(BytesIO(code.encode("utf-8")).readline)
    for toknum, tokval, a, b, c in g:

        if (
            toknum == COMMENT
            and tokval.startswith("# transonic ")
            and "import" in tokval
        ):
            imports.append(tokval.split("# transonic ", 1)[1])

        if toknum == NAME and tokval == "use_block":
            has_to_find_name_block = True
            has_to_find_signatures = True
            has_to_find_code_block = "after use_block"

        if has_to_find_name_block and toknum == STRING:
            name_block = eval(tokval)
            has_to_find_name_block = False
            blocks.append(name_block)

        if toknum == COMMENT and tokval.startswith("# transonic def "):
            in_def = True
            signature_func = tokval.split("# transonic def ", 1)[1]
            name_func = signature_func.split("(")[0]
            functions.add(name_func)

            if name_func not in signatures_func:
                signatures_func[name_func] = []

            if ")" in tokval:
                in_def = False
                signatures_func[name_func].append(signature_func)

        if in_def:
            if toknum == COMMENT:
                if "# transonic def " in tokval:
                    tokval = tokval.split("(", 1)[1]
                signature_func += tokval.replace("#", "").strip()
                if ")" in tokval:
                    in_def = False
                    signatures_func[name_func].append(signature_func)

        if has_to_find_signatures and toknum == COMMENT:
            if tokval.startswith("# transonic block "):
                in_signature = True
                signature = tokval.split("# transonic block ", 1)[1]
            elif in_signature:
                signature += tokval[1:].strip()
                if ")" in tokval and "-> (" not in tokval or tokval.endswith(")"):
                    in_signature = False

                    if name_block not in signatures_blocks:
                        signatures_blocks[name_block] = []
                    signatures_blocks[name_block].append(signature)

        if has_to_find_code_block == "in block":

            if toknum == DEDENT:
                logger.debug(
                    f"code_blocks[name_block]: {code_blocks[name_block]}"
                )

                code_blocks[name_block] = untokenize(code_blocks[name_block])
                has_to_find_code_block = False
            else:
                code_blocks[name_block].append((toknum, tokval))

        if has_to_find_code_block == "after use_block" and toknum == INDENT:
            has_to_find_code_block = "in block"
            code_blocks[name_block] = []

        # logger.debug((tok_name[toknum], tokval))

    return (
        blocks,
        signatures_blocks,
        code_blocks,
        functions,
        signatures_func,
        imports,
    )


def get_code_functions(code: str, func_names: Iterable[str]):
    """Get the code of function from a path and function names"""

    indent = 0
    in_def = False
    codes = {}

    g = tokenize(BytesIO(code.encode("utf-8")).readline)
    for toknum, tokval, a, b, c in g:

        if toknum == INDENT:
            indent += 1

        if toknum == DEDENT:
            indent -= 1

        if in_def == "def" and tokval in func_names:
            in_def = tokval
            codes[in_def] = [(NAME, "def")]

        if indent == 0 and toknum == NAME and tokval == "def":
            in_def = "def"

        if in_def and in_def != "def":

            if toknum == COMMENT and tokval.startswith("# pythran"):
                continue

            if indent == 0 and toknum == DEDENT:
                codes[in_def] = untokenize(codes[in_def])
                in_def = False
            else:
                codes[in_def].append((toknum, tokval))

        logger.debug((indent, tok_name[toknum], tokval))

    return codes


def get_codes_from_functions(functions: dict):
    """create a dict {name: code} from {name: function}

    """
    codes = {}

    for name, func in functions.items():
        code = get_source_without_decorator(func)
        codes[name] = code

    return codes


def make_pythran_code_functions(
    functions: Iterable[str], signatures_func: dict, codes_functions: dict
):
    """Create the pythran code for all functions

    Parameters
    ----------

    functions : list

    signatures_func : dict

    codes_functions: dict

    """
    code_pythran = ""

    for name_func in functions:

        try:
            signatures = signatures_func[name_func]
        except KeyError:
            logger.warning("No Pythran signature for function " + name_func)
            continue

        if not signatures:
            raise RuntimeError("No Pythran signature for function " + name_func)

        for signature in signatures:
            code_pythran += f"# pythran export {signature}\n"

        code = codes_functions[name_func]

        code_pythran += f"\n{code}\n\n"

    return code_pythran


def produce_code_class(cls, jit=False):
    pythran_code = ""
    for key, value in cls.__dict__.items():
        if hasattr(value, "__transonic__") and value.__transonic__ in (
            "trans_def_method",
            "jit_method",
        ):
            pythran_code += produce_code_class_func(cls, key, jit)
    return pythran_code


def produce_new_code_method(cls, func):

    src = get_source_without_decorator(func)

    tokens = []
    attributes = set()

    using_self = False

    g = tokenize(BytesIO(src.encode("utf-8")).readline)
    for toknum, tokval, a, b, c in g:
        logger.debug((tok_name[toknum], tokval))

        if using_self == "self":
            if toknum == OP and tokval == ".":
                using_self = tokval
                continue
            elif toknum == OP and tokval in (",", ")"):
                tokens.append((NAME, "self"))
                using_self = False
            else:
                raise NotImplementedError(
                    f"self{tokval} not supported by Transonic"
                )

        if using_self == ".":
            if toknum == NAME:
                using_self = False
                tokens.append((NAME, "self_" + tokval))
                attributes.add(tokval)
                continue
            else:
                raise NotImplementedError

        if toknum == NAME and tokval == "self":
            using_self = "self"
            continue

        tokens.append((toknum, tokval))

    attributes = sorted(attributes)

    attributes_self = ["self_" + attr for attr in attributes]

    index_self = tokens.index((NAME, "self"))

    tokens_attr = []
    for ind, attr in enumerate(attributes_self):
        tokens_attr.append((NAME, attr))
        tokens_attr.append((OP, ","))

    if tokens[index_self + 1] == (OP, ","):
        del tokens[index_self + 1]

    tokens = tokens[:index_self] + tokens_attr + tokens[index_self + 1 :]

    index_func_name = tokens.index((NAME, func.__name__))
    cls_name = cls.__name__
    name_new_func = f"__for_method__{cls_name}__{func.__name__}"
    tokens[index_func_name] = (NAME, name_new_func)

    new_code = untokenize(tokens).decode("utf-8")

    return new_code, attributes, name_new_func


def produce_code_class_func(cls, func_name, jit=False):

    func = cls.__dict__[func_name]

    if jit:
        func = func.func

    new_code, attributes, name_new_func = produce_new_code_method(cls, func)

    try:
        cls_annotations = cls.__annotations__
    except AttributeError:
        cls_annotations = {}

    if jit:
        types_attrs = [
            cls_annotations[attr]
            for attr in attributes
            if attr in cls_annotations
        ]
    else:
        for attr in attributes:
            if attr not in cls_annotations:
                raise NotImplementedError(
                    f"self.{attr} used but {attr} not in class annotations"
                )
        types_attrs = [cls_annotations[attr] for attr in attributes]

    signature = inspect.signature(func)
    types_func = [param.annotation for param in signature.parameters.values()][1:]
    types_pythran = types_attrs + types_func

    pythran_signatures = "\n"

    try:
        types_string_signatures = compute_pythran_types_from_valued_types(
            types_pythran
        )
    except ValueError:
        if jit:
            types_string_signatures = []
        else:
            raise

    for types_string_signature in types_string_signatures:
        pythran_signatures += (
            "# pythran export "
            + name_new_func
            + "("
            + ", ".join(types_string_signature)
            + ")\n"
        )

    if jit:
        new_code = "from transonic import jit\n\n@jit\n" + new_code

    python_code = pythran_signatures + "\n" + new_code

    str_self_dot_attributes = ", ".join("self." + attr for attr in attributes)
    args_func = list(signature.parameters.keys())[1:]
    str_args_func = ", ".join(args_func)

    str_args_value_func = ""
    for param, value in signature.parameters.items():
        if param == "self":
            continue
        elif value.default is value.empty:
            str_args_value_func += f"{param}, "
        else:
            str_args_value_func += f"{param}={value.default}, "

    str_args_value_func = str_args_value_func.rstrip(", ")

    if jit:
        name_new_method = f"__new_method__{cls.__name__}__{func_name}"
        python_code += (
            f"\ndef {name_new_method}"
            f"(self, {str_args_value_func}):\n"
            f"    return {name_new_func}({str_self_dot_attributes}, {str_args_func})"
            "\n"
        )
    else:
        name_var_code_new_method = (
            f"__code_new_method__{cls.__name__}__{func_name}"
        )
        python_code += (
            f"\n# pythran export {name_var_code_new_method}\n"
            f'\n{name_var_code_new_method} = """\n\n'
            f"def new_method(self, {str_args_value_func}):\n"
            f"    return pythran_func({str_self_dot_attributes}, {str_args_func})"
            '\n\n"""\n'
        )

    if black:
        python_code = black.format_str(python_code, line_length=82)

    return python_code


def make_pythran_code(path_py: Path):
    """Create a pythran code from a Python file"""

    with open(path_py) as file:
        code = file.read()

    namespace = None

    module_name = find_module_name_from_path(path_py)

    if "# TRANSONIC_NO_IMPORT" not in code:
        # we have to import the module!
        transonic.aheadoftime.is_transpiling = True
        try:
            namespace = run_path(fspath(path_py))
        except ImportError:
            sys.path.insert(0, "")
            try:
                namespace = run_module(module_name)
            except ImportError:
                logger.error(
                    f"transonic was unable to import module {module_name}: "
                    "no Pythran file created. "
                    "You could add '# TRANSONIC_NO_IMPORT' "
                    "in the module if needed..."
                    "You could mock modules by using the argument mocked_modules to "
                    "the function make_backend_files."
                )
                raise
            finally:
                del sys.path[0]
        transonic.aheadoftime.is_transpiling = False

    (
        blocks,
        signatures_blocks,
        code_blocks,
        functions,
        signatures_func,
        imports,
    ) = parse_py_code(code)

    if logger.is_enable_for("debug"):
        logger.debug(
            f"""
blocks: {blocks}\n
signatures_blocks: {signatures_blocks}\n
code_blocks:  {code_blocks}\n
functions: {functions}\n
signatures_func: {signatures_func}\n
imports: {imports}\n"""
        )

    code_pythran = ""
    if imports:
        code_pythran += "\n" + "\n".join(imports) + "\n"

    if module_name in transonic.aheadoftime.modules:
        ts = transonic.aheadoftime.modules[module_name]
        ts._make_signatures_from_annotations()
        functions = ts.functions.keys()
        signatures_func_annot = ts.signatures_func
        codes_functions = get_codes_from_functions(ts.functions)

        # merge signatures introduced by type annotations and by Pythran commands
        signatures_func_all = signatures_func.copy()
        for name_func, signatures in signatures_func_annot.items():
            if name_func not in signatures_func_all:
                signatures_func_all[name_func] = []
            signatures_func_all[name_func].extend(signatures)

        for cls in ts.classes.values():
            code_pythran += produce_code_class(cls)

        for func in ts.included_functions:
            code_pythran += "\n" + get_source_without_decorator(func) + "\n"

    else:
        if "# TRANSONIC_NO_IMPORT" not in code:
            logger.warning(
                "module_name not in transonic.aheadoftime.modules\n"
                f"module_name = {module_name}\n"
                f"path_py = {path_py}"
            )

        codes_functions = get_code_functions(code, functions)
        signatures_func_all = signatures_func

    code_pythran += make_pythran_code_functions(
        functions, signatures_func_all, codes_functions
    )

    # blocks...

    # we check that some types correspond to transonic types
    # we can do that only if the module has been imported

    types_variables_blocks = {}
    return_block = {}
    for name_block, list_str_types_variables in signatures_blocks.items():
        types_variables_blocks[name_block] = []

        for str_types_variables in list_str_types_variables:
            types_variables = {}
            lines = str_types_variables.split("(", 1)[1].split(")")[0].split(";")
            for line in lines:
                type_, str_variables = line.strip().split(" ", 1)

                if namespace is not None and type_ in namespace:
                    type_ = namespace[type_]
                else:
                    try:
                        type_ = eval(type_)
                    except (SyntaxError, TypeError):
                        pass

                variables = [
                    variable.replace(",", "").strip()
                    for variable in str_variables.split(",")
                ]

                types_variables[type_] = variables

            types_variables_blocks[name_block].append(types_variables)

            if list_str_types_variables and "->" in str_types_variables:
                return_block[name_block] = str_types_variables.split("->", 1)[1]

    for name_block, list_types_variables in tuple(types_variables_blocks.items()):
        new_list_types_variables = []
        for types_variables in list_types_variables:
            sequence_types = compute_pythran_types_from_valued_types(
                types_variables.keys()
            )
            variabless = types_variables.values()
            # print("variabless", variabless)
            for types in sequence_types:
                new_types_variables = {}
                for type_, variables in zip(types, variabless):
                    if type_ not in new_types_variables:
                        new_types_variables[type_] = variables
                    else:
                        new_types_variables[type_] = (
                            new_types_variables[type_] + variables
                        )
                new_list_types_variables.append(new_types_variables)

        types_variables_blocks[name_block] = new_list_types_variables

    # print("types_variables_blocks", types_variables_blocks)

    variables_types_block = {}
    for name_block, list_types_variables in types_variables_blocks.items():
        variables_types_block[name_block] = []
        for types_variables in list_types_variables:
            variables_types = {}
            for type_, variables in types_variables.items():
                for variable in variables:
                    variables_types[variable] = type_
                variables_types

            variables_types_block[name_block].append(variables_types)

    # print("variables_types_block", variables_types_block)

    arguments_blocks = {}

    for name_block, list_variables_types in variables_types_block.items():
        # add "pythran export" for blocks

        # print("list_variables_types", list_variables_types)

        tmp = []

        for variables_types in list_variables_types:
            if variables_types in tmp:
                continue
            tmp.append(variables_types)
            str_types = variables_types.values()
            str_variables = ", ".join(str_types)
            code_pythran += f"# pythran export {name_block}({str_variables})\n"

        # add code for blocks
        variables = list_variables_types[0].keys()
        arguments_blocks[name_block] = list(variables)

        str_variables = ", ".join(variables)

        code_pythran += f"\ndef {name_block}({str_variables}):\n"

        code_block = code_blocks[name_block]

        code_pythran += 4 * " " + code_block.replace("\n", "\n" + 4 * " ") + "\n"

        if name_block in return_block:
            code_pythran += f"    return {return_block[name_block]}\n"

    if arguments_blocks:
        code_pythran += "# pythran export arguments_blocks\n"
        code_pythran += "arguments_blocks = " + str(arguments_blocks) + "\n"

    if code_pythran:
        code_pythran += (
            "\n# pythran export __transonic__\n"
            f"__transonic__ = ('{transonic.__version__}',)"
        )

    if black:
        code_pythran = black.format_str(code_pythran, line_length=82)

    return code_pythran


def make_pythran_file(
    path_py: Path,
    force=False,
    log_level=None,
    mocked_modules: Optional[Iterable] = None,
):
    """Create a Python file from a Python file (if necessary)"""
    if log_level is not None:
        logger.set_level(log_level)

    path_py = Path(path_py)

    if not path_py.exists():
        raise FileNotFoundError(f"Input file {path_py} not found")

    if path_py.absolute().parent.name == "__pythran__":
        logger.debug(f"skip file {path_py}")
        return
    if not path_py.name.endswith(".py"):
        raise ValueError(
            "transonic only processes Python file. Cannot process {path_py}"
        )

    path_dir = path_py.parent / "__pythran__"
    path_pythran = path_dir / path_py.name

    if not has_to_build(path_pythran, path_py) and not force:
        logger.warning(f"File {path_pythran} already up-to-date.")
        return

    with mock_modules(mocked_modules):
        code_pythran = make_pythran_code(path_py)

    if not code_pythran:
        return

    if path_pythran.exists() and not force:
        with open(path_pythran) as file:
            code_pythran_old = file.read()

        if code_pythran_old == code_pythran:
            logger.warning(f"Code in file {path_pythran} already up-to-date.")
            return

    logger.debug(f"code_pythran:\n{code_pythran}")

    path_dir.mkdir(exist_ok=True)

    with open(path_pythran, "w") as file:
        file.write(code_pythran)

    logger.info(f"File {fspath(path_pythran)} written")

    return path_pythran


def make_backend_files(
    paths: Iterable[Path],
    force=False,
    log_level=None,
    mocked_modules: Optional[Iterable] = None,
    backend=None,
):
    """Create Pythran files from a list of Python files"""

    assert backend is None

    if log_level is not None:
        logger.set_level(log_level)

    paths_out = []
    for path in paths:
        path_out = make_pythran_file(
            path, force=force, mocked_modules=mocked_modules
        )
        if path_out:
            paths_out.append(path_out)

    if paths_out:
        nb_files = len(paths_out)
        if nb_files == 1:
            conjug = "s"
        else:
            conjug = ""

        logger.warning(
            f"{nb_files} files created or updated need{conjug}"
            " to be pythranized"
        )

    return paths_out


class _MyMock(Mock):
    @classmethod
    def __getattr__(cls, name):
        return Mock()


@contextmanager
def mock_modules(modules):
    """Context manager to mock modules

    Examples
    --------

    .. code::

        with mock_modules(("h5py", "reikna.fft", "reikna.transformations")):
            code_pythran = make_pythran_code(path_py)

    """
    if modules is not None:
        modules = set(modules)
        modules.difference_update(set(sys.modules.keys()))
        sys.modules.update((mod_name, _MyMock()) for mod_name in modules)

    try:
        yield None
    finally:
        if modules is not None:
            for mod_name in modules:
                sys.modules.pop(mod_name)






import numpy as np

from transonic import Transonic, Type, NDim, Array

T = Type(float, complex)
N = NDim(1, 2)
A = Array[T, N]
A1 = Array[T, N + 1]

ts = Transonic()


class MyClass:
    def __init__(self, a, b):
        self.a = a
        self.b = b

    def compute(self, n):

        a = self.a
        b = self.b

        if ts.is_transpiled:
            result = ts.use_block("block0")
        else:
            # transonic block (
            #     A a; A1 b;
            #     int n
            # ) -> result

            # transonic block (
            #     int[:] a, b;
            #     float n
            # ) -> result

            result = a ** 2 + b.mean() ** 3 + n

        return result


if __name__ == "__main__":

    shape = 100, 100
    a = np.random.rand(*shape)
    b = np.random.rand(*shape)

    obj = MyClass(a, b)

    obj.compute(10)

    if ts.is_transpiled:
        ret = obj.compute(10)
        ts.is_transpiled = False
        ret1 = obj.compute(10)
        ts.is_transpiled = True
        assert np.allclose(ret, ret1)
        print("allclose OK")



"""
Not yet implemented...

Many things can be expressed in Pythran specifications (see
https://pythran.readthedocs.io/en/latest/MANUAL.html#concerning-pythran-specifications),
in particular stride arrays and partial shapes...

We could also express these concepts in strings, mainly following Pythran...

"""

from transonic import boost, Type, NDim, Shape, Array

T = Type(int, float)

# here the shape of the array is only defined with the ShapeVar
A = Array[T, Shape("[3, :]", "[3, :, :]", "[::, ::]", "[::, ::, ::]")]


@boost
def compute(a: A, b: A, c: T):
    return a + b


# if there is a NDimVar, we can use the ellipsis
A1 = Array[T, NDim(1, 3), Shape("[3, ...]", "[::, ...]")]


@boost
def compute1(a: A1, b: A1, c: T):
    return c * (a + b)



import numpy as np

from transonic import boost

# transonic import numpy as np


def mylog(arr):
    return np.log(arr)

# transonic def func(float[][], float[][])
# transonic def func(int[][], float[][])


@boost
def func(a, b):
    return (a * mylog(b)).max()


from transonic import boost, include


@include
def add(a, b):
    return a + b


@boost
class MyClass:
    attr0: float
    attr1: float

    def __init__(self, arg):
        self.attr0 = self.attr1 = 2 * float(arg)

    @boost
    def compute(self, number: int):
        result = 0.0
        for _ in range(number):
            result += add(self.attr0, self.attr1)



import importlib
import unittest
import os
import time

try:
    import pythran
except ImportError:
    pythran = False


from .transpiler import make_pythran_file
from .util import has_to_compile_at_import, ext_suffix, name_ext_from_path_backend
from .aheadoftime import modules
from . import mpi

module_name = "transonic.for_test_init"


class TestsInit(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        cls.path_for_test = mpi.Path(__file__).parent / "for_test_init.py"

        assert cls.path_for_test.exists()

        cls.path_pythran = path_pythran = (
            cls.path_for_test.parent / "__pythran__" / cls.path_for_test.name
        )
        cls.path_ext = path_pythran.with_name(
            name_ext_from_path_backend(path_pythran)
        )

    @classmethod
    def tearDownClass(cls):
        # cls.path_pythran.unlink()
        if cls.path_ext.exists():
            cls.path_ext.unlink()

        path_ext = cls.path_ext.with_suffix(ext_suffix)
        if path_ext.exists():
            path_ext.unlink()

        try:
            os.environ.pop("TRANSONIC_COMPILE_AT_IMPORT")
        except KeyError:
            pass

        print(mpi.rank, "end tearDownClass")

    def test_transonified(self):

        print(mpi.rank, "start test", flush=1)

        try:
            os.environ.pop("TRANSONIC_COMPILE_AT_IMPORT")
        except KeyError:
            pass

        try:
            del modules[module_name]
        except KeyError:
            pass

        assert not has_to_compile_at_import()

        print(mpi.rank, "before if self.path_pythran.exists()", flush=1)

        if self.path_pythran.exists():

            print(mpi.rank, "before self.path_pythran.unlink()", flush=1)

            self.path_pythran.unlink()

        print(mpi.rank, "before make_pythran_file(self.path_for_test)", flush=1)
        if mpi.rank == 0:
            make_pythran_file(self.path_for_test)

        print(mpi.rank, "after make_pythran_file(self.path_for_test)", flush=1)
        mpi.barrier()

        from . import for_test_init

        importlib.reload(for_test_init)

        assert self.path_pythran.exists()
        assert for_test_init.ts.is_transpiled

        for_test_init.func(1, 3.14)
        for_test_init.func1(1.1, 2.2)
        for_test_init.check_class()

    @unittest.skipIf(
        not pythran, "Pythran is required for TRANSONIC_COMPILE_AT_IMPORT"
    )
    def test_pythranize(self):

        os.environ["TRANSONIC_COMPILE_AT_IMPORT"] = "1"

        try:
            del modules[module_name]
        except KeyError:
            pass

        assert has_to_compile_at_import()

        if self.path_pythran.exists():
            self.path_pythran.unlink()

        if self.path_ext.exists():
            self.path_ext.unlink()

        path_ext = self.path_ext.with_suffix(ext_suffix)
        if path_ext.exists():
            path_ext.unlink()

        from . import for_test_init

        if not for_test_init.ts.is_compiling:
            importlib.reload(for_test_init)

        assert module_name in modules, modules

        assert self.path_pythran.exists()

        ts = for_test_init.ts

        assert ts.is_transpiled
        assert ts.is_compiling
        assert not ts.is_compiled

        for_test_init.func(1, 3.14)
        for_test_init.func1(1.1, 2.2)

        while not ts.is_compiled:
            time.sleep(0.1)
            for_test_init.func(1, 3.14)
            for_test_init.func1(1.1, 2.2)

        assert not ts.is_compiling
        assert ts.is_compiled

        for_test_init.func(1, 3.14)
        for_test_init.func1(1.1, 2.2)
        for_test_init.check_class()



import beniget
import gast as ast
import astunparse

from capturex import CaptureX

def dump(node):
    print(astunparse.dump(node))


with open("issue7_method.py") as file:
    code = file.read()

module = ast.parse(code)

duc = beniget.DefUseChains()
duc.visit(module)

ucc = beniget.UseDefChains(duc)

ancestors = beniget.Ancestors()
ancestors.visit(module)

boost = [d for d in duc.locals[module] if d.name() == "boost"][0]

chain = duc.chains[boost.node]
users = chain.users()

class_defs = []
function_defs = []

for user in users:
    func_or_class_def = ancestors.parent(user.node)
    if isinstance(func_or_class_def, ast.ClassDef):
        class_defs.append(func_or_class_def)
    elif isinstance(func_or_class_def, ast.FunctionDef):
        function_defs.append(func_or_class_def)


for fdef in function_defs:
    for node in fdef.decorator_list:
        if node.id == "boost":
            fdef.decorator_list.remove(node)

    capturex = CaptureX(module, fdef)

    capturex.visit(fdef)

    for node in capturex.external:
        # print(astunparse.dump(node))
        print(astunparse.unparse(node).strip())




import numpy as np

from transonic import boost

# TRANSONIC_NO_IMPORT

# transonic import numpy as np

# transonic def func(
#  float[][],
#  float[][]
# )
# transonic def func(int[][], float[][])


@boost
def func(a, b):
    return (a * np.log(b)).max()



import numpy as np

from transonic import Transonic

ts = Transonic()


class MyClass:
    def __init__(self, a, b):
        self.a = a
        self.b = b

    def compute(self, n):

        a = self.a
        b = self.b

        if ts.is_transpiled:
            result = ts.use_block("block0")
        else:
            # transonic block (
            #     float[][] a, b;
            #     int n
            # ) -> result
            # blabla

            # blibli

            # transonic block (
            #     float[][][] a, b;
            #     int n
            # ) -> result
            result = np.zeros_like(a)
            for _ in range(n):
                result += a ** 2 + b ** 3

        a = result

        if ts.is_transpiled:
            result = ts.use_block("block1")
        else:
            # transonic block (
            #     float[][] a, b;
            #     int n
            # ) -> (result)
            # blabla

            # blibli

            # transonic block (
            #     float[][][] a, b;
            #     int n
            # ) -> (result)
            result = np.zeros_like(a)
            for _ in range(n):
                result += a ** 2 + b ** 3

        return result


if __name__ == "__main__":

    shape = 2, 2
    a = np.random.rand(*shape)
    b = np.random.rand(*shape)

    obj = MyClass(a, b)

    ret0 = obj.compute(10)

    print("(is_transpiled, is_compiling, is_compiled)", (ts.is_transpiled, ts.is_compiling, ts.is_compiled))

    if ts.is_transpiled:
        ret = obj.compute(10)
        assert np.allclose(ret, ret0), (ret - ret0)
        ts.is_transpiled = False
        ret1 = obj.compute(10)
        ts.is_transpiled = True
        assert np.allclose(ret, ret1), (ret - ret1)
        print("allclose OK")



"""
See https://bitbucket.org/fluiddyn/fluidpythran/issues/29/

"""
from transonic import with_blocks, block


def non_pythranizable(arg):
    """represent a function that can not be compiled by Pythran"""
    return arg


@with_blocks
def func0(arg: int):

    a = non_pythranizable(arg)

    with block(a=float):
        return a + arg


@with_blocks
def func1(arg: int):

    a = non_pythranizable(arg)

    with block(a=float):
        tmp = a + arg

    return non_pythranizable(tmp)



from ._version import __version__
from .justintime import jit, set_compile_jit

try:
    from ._path_data_tests import path_data_tests
except ImportError:
    pass

from .annotation import Array, NDim, Type, Shape, Union

from .aheadoftime import Transonic, boost, make_signature, include

from .util import set_compile_at_import
from .pythranizer import wait_for_all_extensions

__all__ = [
    "__version__",
    "Transonic",
    "boost",
    "include",
    "make_signature",
    "path_data_tests",
    "Array",
    "NDim",
    "Type",
    "Shape",
    "Union",
    "jit",
    "set_compile_jit",
    "set_compile_at_import",
    "wait_for_all_extensions",
]



import numpy as np

from transonic import boost

# transonic def func(float[][], float[][])
# transonic def func(int[][], float[][])


@boost
def func(a: float, b: float):
    return (a * np.log(b)).max()


@boost
def func1(a: int, b: float):
    return a * np.cos(b)



import numpy as np
import transonic as ts
from transonic import Type, NDim, Array, Union

T = Type(int, np.complex128)
N = NDim(1, 3)

A = Array[T, N]
A1 = Array[np.float32, N + 1]

A3d = Array[np.float32, "3d"]
N1 = NDim(4, 5)
N1 = NDim(4, 5)

T = Type(int, np.complex128)

myconst = 0

@ts.boost
def compute(a: A, b: A, c: T, d: Union[A, A1], e: str):
    print(e)
    tmp = a + b + myconst
    return tmp





from transonic import Transonic

ts = Transonic()


class MyClass:
    def _time_step_RK2(self):
        dt = self.deltat
        diss, diss2 = self.exact_linear_coefs.get_updated_coefs()

        compute_tendencies = self.sim.tendencies_nonlin
        state_spect = self.sim.state.state_spect

        tendencies_n = compute_tendencies()

        state_spect_n12 = self._state_spect_tmp

        if ts.is_transpiled:
            ts.use_block("rk2_step0")
        else:
            # transonic block (
            #     complex128[][][] state_spect_n12, state_spect,
            #                      tendencies_n;
            #     float64[][] diss2;
            #     float dt
            # )
            state_spect_n12[:] = (state_spect + dt / 2 * tendencies_n) * diss2

        print("hello!")


import numpy as np
import transonic as ts
from transonic import Type, NDim, Array

T = Type("T")
T1 = Type("T1")
N = NDim("N")

A = Array[T, N]
A1 = Array[T1, N + 1]

# for coverage
assert repr(N - 1) == "N - 1"
print(repr(A1))


@ts.boost
def compute(a: A, b: A1, c: T, d: A, e: str):
    print(e)
    tmp = a + b
    return tmp


for dtype in [int, np.complex128]:
    for ndim in [1, 3]:
        ts.make_signature(compute, T=dtype, N=ndim, T1=float)


if __name__ == "__main__":
    print(__name__)



import numpy as np
from transonic import Type, NDim, Array, boost

T = Type(int, np.complex128)
N = NDim(1, 3)

A = Array[T, N]
A1 = Array[np.float32, N + 1]


@boost
def compute(a: A, b: A, c: T, d: A1, e: str):
    print(e)
    tmp = a + b
    return tmp



import numpy as np

from transonic import boost
import transonic as ts

myconst0 = 0
myconst1 = 2*myconst0

# transonic def func(
#  float[][],
#  float[][]
# )
# transonic def func(int[][], float[][])


@boost
def func(a, b):
    return myconst1 * (a * np.log(b)).max()


# transonic def func1(int, float)


@ts.boost
def func1(a, b):
    return (a * np.cos(b)).max()


@boost
class Foo:
    @ts.boost
    def meth_foo(self):
        return func1(0, 1)


@ts.boost
class Bar:
    @boost
    def meth_bar(self):
        pass




from runpy import run_module

import numpy as np

inp = np.ones(2)


def check(module_name, should_return=None):
    mod = run_module(module_name)
    trans = mod["Transmitter"](1.)
    assert trans.__call__.__doc__ == "My docstring", trans.__call__.__doc__
    assert trans.__call__.__name__ == "__call__"

    result = trans(inp)

    if should_return is not None:
        assert np.allclose(result, should_return)

    return result


original = check("transmitter")
check("try0", original)
check("try1", original)
check("try2", original)
check("try3", original)
check("try4", original)



import os
from pathlib import Path
import runpy
import shutil
from contextlib import suppress

import pytest
from . import path_data_tests
from .dist import make_backend_files
from .mpi import nb_proc

cwd = Path.cwd().absolute()
setup_dir = path_data_tests / "test_packaging"


def setup_module():
    os.chdir(setup_dir)
    trasonic_src_paths = [setup_dir / "add.py"]
    make_backend_files(trasonic_src_paths)


@pytest.mark.skipif(nb_proc > 1, reason="No build_ext in MPI")
def test_buildext():
    os.chdir(setup_dir)
    runpy.run_path(str(setup_dir / "setup.py"))


def teardown_module():
    os.chdir(cwd)
    for namedir in ("build", "__pythran__", "__pycache__"):
        with suppress(FileNotFoundError):
            shutil.rmtree(setup_dir / namedir)

    to_remove = list(setup_dir.glob("*.h")) + list(setup_dir.glob("*.so"))
    for path in to_remove:
        os.remove(path)



"""Cached JIT compilation
=========================

User API
--------

.. autofunction:: jit

.. autofunction:: set_compile_jit

Internal API
------------

.. autoclass:: ModuleJIT
   :members:
   :private-members:

.. autofunction:: _get_module_jit

.. autoclass:: JIT
   :members:
   :private-members:

.. autofunction:: make_pythran_type_name

Notes
-----

Serge talked about @jit (see https://gist.github.com/serge-sans-paille/28c86d2b33cd561ba5e50081716b2cf4)

It's indeed a good idea!

With "# transonic import" and @include the implementation isn't
too complicated.

- At import time, we create one .py file per jit function.

- At run time, we create (and complete when needed) a corresponding
  .pythran file with signature(s).

  The jit decorator:

  * at the first call, get the types, create the .pythran file and call
    Pythran.

  * then, try to call the pythran function and if it fails with
    a Pythran TypeError, correct the .pythran file and recompile.

Note: During the compilation (the "warmup" of the JIT), the Python function is
used.

"""

import inspect
import itertools
import os
import sys
import time
from distutils.util import strtobool
from functools import wraps

try:
    import numpy as np
except ImportError:
    np = None

from .pythranizer import compile_extension, ext_suffix

from .util import (
    get_module_name,
    has_to_build,
    get_source_without_decorator,
    path_root,
    get_info_from_ipython,
    make_hex,
    has_to_compile_at_import,
    import_from_path,
    pythran,
    is_method,
)
from .annotation import make_signatures_from_typehinted_func, normalize_type_name

from .aheadoftime import TransonicTemporaryJITMethod

from .compat import open
from . import mpi
from .log import logger
from .config import has_to_replace

modules = {}


path_jit = mpi.Path(path_root) / "__jit__"
if mpi.rank == 0:
    path_jit.mkdir(parents=True, exist_ok=True)
mpi.barrier()

_COMPILE_JIT = strtobool(os.environ.get("TRANSONIC_COMPILE_JIT", "True"))


def set_compile_jit(value):
    global _COMPILE_JIT
    _COMPILE_JIT = value


class ModuleJIT:
    """Representation of a module using jit"""

    def __init__(self, frame=None):

        if frame is None:
            frame = inspect.stack()[1]

        self.filename = frame.filename
        if self.filename.startswith("<ipython-"):
            self.is_dummy_file = True
            self._ipython_src, self.filename = get_info_from_ipython()
            self.module_name = self.filename
        else:
            self.is_dummy_file = False
            self.module_name = get_module_name(frame)
        modules[self.module_name] = self
        self.used_functions = {}
        self.jit_functions = {}

    def record_used_function(self, func, names):
        if isinstance(names, str):
            names = (names,)

        for name in names:
            if name in self.used_functions:
                self.used_functions[name].append(func)
            else:
                self.used_functions[name] = [func]

    def get_source(self):
        if self.is_dummy_file:
            return self._ipython_src
        try:
            mod = sys.modules[self.module_name]
        except KeyError:
            with open(self.filename) as file:
                return file.read()
        else:
            return inspect.getsource(mod)


def _get_module_jit(index_frame: int = 2):
    """Get the ModuleJIT instance corresponding to the calling module

    Parameters
    ----------

    index_frame : int

      Index (in :code:`inspect.stack()`) of the frame to be selected

    """
    try:
        frame = inspect.stack()[index_frame]
    except IndexError:
        logger.error(
            f"index_frame {index_frame}"
            f"{[frame[1] for frame in inspect.stack()]}"
        )
        raise

    module_name = get_module_name(frame)

    if module_name in modules:
        return modules[module_name]
    else:
        return ModuleJIT(frame=frame)


def make_pythran_type_name(obj: object):
    """return the Pythran type name"""
    name = type(obj).__name__
    name = normalize_type_name(name)

    if np and isinstance(obj, np.ndarray):
        name = obj.dtype.name
        if obj.ndim != 0:
            name += "[" + ", ".join([":"] * obj.ndim) + "]"

    if name in ("list", "set", "dict"):
        if not obj:
            raise ValueError(
                f"cannot determine the Pythran type from an empty {name}"
            )

    if name in ("list", "set"):
        item_type = type(obj[0])
        # FIXME: we could check if the iterable is homogeneous...
        name = item_type.__name__ + " " + name

    if name == "dict":
        for key, value in obj.items():
            break
        # FIXME: we could check if the dict is homogeneous...
        name = type(key).__name__ + ": " + type(value).__name__ + " dict"

    return name


def jit(func=None, native=True, xsimd=True, openmp=False):
    """Decorator to record that the function has to be jit compiled

    """
    decor = JIT(native=native, xsimd=xsimd, openmp=openmp)
    if callable(func):
        decor._decorator_no_arg = True
        return decor(func)
    else:
        return decor


class JIT:
    """Decorator used internally by the public jit decorator
    """

    def __init__(self, native=True, xsimd=True, openmp=False):
        self.native = native
        self.xsimd = xsimd
        self.openmp = openmp
        self._decorator_no_arg = False

        self.pythran_func = None
        self.compiling = False
        self.process = None

    def __call__(self, func):

        if not has_to_replace:
            return func

        if is_method(func):
            return TransonicTemporaryJITMethod(
                func, self.native, self.xsimd, self.openmp
            )

        if not pythran:
            return func

        func_name = func.__name__

        if self._decorator_no_arg:
            index_frame = 3
        else:
            index_frame = 2

        mod = _get_module_jit(index_frame)
        mod.jit_functions[func_name] = self
        module_name = mod.module_name

        path_pythran = path_jit / module_name.replace(".", os.path.sep)

        if mpi.rank == 0:
            path_pythran.mkdir(parents=True, exist_ok=True)
        mpi.barrier()

        path_pythran = (path_pythran / func_name).with_suffix(".py")
        path_pythran_header = path_pythran.with_suffix(".pythran")

        if path_pythran.exists():
            if not mod.is_dummy_file and has_to_build(path_pythran, mod.filename):
                has_to_write = True
            else:
                has_to_write = False
        else:
            has_to_write = True

        src = None

        if has_to_write:
            import_lines = [
                line.split("# transonic ")[1]
                for line in mod.get_source().split("\n")
                if line.startswith("# transonic ") and "import" in line
            ]
            src = "\n".join(import_lines) + "\n\n"

            src += get_source_without_decorator(func)

            if func_name in mod.used_functions:
                functions = mod.used_functions[func_name]
                for function in functions:
                    src += "\n" + get_source_without_decorator(function)

            if path_pythran.exists() and mpi.rank == 0:
                with open(path_pythran) as file:
                    src_old = file.read()
                if src_old == src:
                    has_to_write = False

            if has_to_write and mpi.rank == 0:
                logger.debug(f"write code in file {path_pythran}")
                with open(path_pythran, "w") as file:
                    file.write(src)
                    file.flush()

        if src is None and mpi.rank == 0:
            with open(path_pythran) as file:
                src = file.read()

        hex_src = None
        name_mod = None
        if mpi.rank == 0:
            # hash from src (to produce the extension name)
            hex_src = make_hex(src)
            name_mod = ".".join(
                path_pythran.absolute()
                .relative_to(path_root)
                .with_suffix("")
                .parts
            )

        hex_src = mpi.bcast(hex_src)
        name_mod = mpi.bcast(name_mod)

        def pythranize_with_new_header(arg_types="no types"):

            # Include signature comming from type hints
            signatures = make_signatures_from_typehinted_func(func)
            exports = set("export " + signature for signature in signatures)

            if arg_types != "no types":
                export_new = "export {}({})".format(
                    func_name, ", ".join(arg_types)
                )
                if export_new not in exports:
                    exports.add(export_new)

            if not exports:
                return

            try:
                path_pythran_header_exists = path_pythran_header.exists()
            except TimeoutError:
                raise RuntimeError(
                    f"A MPI communication in Transonic failed when compiling "
                    f"function {func}. This usually arises when a jitted "
                    "function has to be compiled in MPI and is only called "
                    f"by one process (rank={mpi.rank})."
                )

            if path_pythran_header_exists:
                # get the old signature(s)

                exports_old = None
                if mpi.rank == 0:
                    with open(path_pythran_header) as file:
                        exports_old = [
                            export.strip() for export in file.readlines()
                        ]
                exports_old = mpi.bcast(exports_old)

                # FIXME: what do we do with the old signatures?
                exports.update(exports_old)

            header = "\n".join(sorted(exports)) + "\n"

            mpi.barrier()
            if mpi.rank == 0:
                logger.debug(
                    f"write Pythran signature in file {path_pythran_header} with types\n{arg_types}"
                )
                with open(path_pythran_header, "w") as file:
                    file.write(header)
                    file.flush()

            # compute the new path of the extension
            hex_header = make_hex(header)
            # if mpi.nb_proc > 1:
            #     hex_header0 = mpi.bcast(hex_header)
            #     assert hex_header0 == hex_header
            name_ext_file = (
                func_name + "_" + hex_src + "_" + hex_header + ext_suffix
            )
            self.path_extension = path_pythran.with_name(name_ext_file)

            self.compiling = True

            self.process = compile_extension(
                path_pythran,
                name_ext_file,
                native=self.native,
                xsimd=self.xsimd,
                openmp=self.openmp,
            )

        ext_files = None
        if mpi.rank == 0:
            glob_name_ext_file = func_name + "_" + hex_src + "_*" + ext_suffix
            ext_files = list(
                mpi.PathSeq(path_pythran).parent.glob(glob_name_ext_file)
            )
        ext_files = mpi.bcast(ext_files)

        if not ext_files:
            if has_to_compile_at_import() and _COMPILE_JIT:
                pythranize_with_new_header()
            self.pythran_func = None
        else:
            path_ext = max(ext_files, key=lambda p: p.stat().st_ctime)
            module_pythran = import_from_path(path_ext, name_mod)
            self.pythran_func = getattr(module_pythran, func_name)

        @wraps(func)
        def type_collector(*args, **kwargs):

            if self.compiling:
                if not self.process.is_alive():
                    self.compiling = False
                    time.sleep(0.1)
                    module_pythran = import_from_path(
                        self.path_extension, name_mod
                    )
                    assert hasattr(module_pythran, "__pythran__")
                    self.pythran_func = getattr(module_pythran, func_name)

            try:
                return self.pythran_func(*args, **kwargs)
            except TypeError:
                # need to compiled or recompile
                pass

            if self.compiling or not _COMPILE_JIT:
                return func(*args, **kwargs)

            arg_types = [
                make_pythran_type_name(arg)
                for arg in itertools.chain(args, kwargs.values())
            ]

            pythranize_with_new_header(arg_types)

            return func(*args, **kwargs)

        return type_collector



import numpy as np

# transonic import numpy as np

from transonic import Transonic, boost, include, Array, Union


# transonic def func(int, float)


@boost
def func(a, b):
    return a + b


@include
def func_tmp(arg):
    return arg ** 2


A = Union[int, Array[int, "1d"]]


@boost
def func2(a: A, b: float):
    return a - func_tmp(b)


ts = Transonic()


def func1(a, b):
    n = 10

    if ts.is_transpiled:
        result = ts.use_block("block0")
    else:
        # transonic block (
        #     float a, b;
        #     int n
        # ) -> (result, a)
        # blabla

        result = 0.0
        for _ in range(n):
            result += a ** 2 + b ** 3


@boost
class Transmitter:

    freq: float

    def __init__(self, freq):
        self.freq = float(freq)

    @boost
    def __call__(self, inp: "float[]"):
        """My docstring"""
        return inp * np.exp(np.arange(len(inp)) * self.freq * 1j)

    @boost
    def other_func(self):
        return 2 * self.freq


def check_class():
    inp = np.ones(2)
    freq = 1.0
    trans = Transmitter(freq)

    def for_check(freq, inp):
        return inp * np.exp(np.arange(len(inp)) * freq * 1j)

    assert np.allclose(trans(inp), for_check(freq, inp))
    assert trans.other_func() == 2 * freq


if __name__ == "__main__":
    check_class()



"""Logging
==========

Defines the transonic logger (variable :code:`logger`).

"""

import logging
from types import MethodType

logger = logging.getLogger("transonic")

# Initialize logging
try:
    # Set a nice colored output
    from colorlog import ColoredFormatter

    formatter = ColoredFormatter(
        "%(log_color)s%(levelname)-8s%(reset)s %(blue)s%(message)s",
        log_colors={
            "DEBUG": "cyan",
            "INFO": "green",
            "WARNING": "yellow",
            "ERROR": "red",
            "CRITICAL": "red",
        },
    )
    stream = logging.StreamHandler()
    stream.setFormatter(formatter)
    logger.addHandler(stream)
except ImportError:
    # No color available, use default config
    logging.basicConfig(format="%(levelname)s: %(message)s")
    logger.info("Disabling color, you really want to install colorlog.")


def _get_level_number(level):
    level = level.lower()
    if level == "info":
        return logging.INFO
    elif level == "debug":
        return logging.DEBUG
    elif level == "warning":
        return logging.WARNING
    else:
        raise ValueError


def set_level(self, level):
    """Set logging level"""
    if not level:
        self.setLevel(0)
        return

    level = _get_level_number(level)

    self.setLevel(level)


def get_level(self):
    return logging.getLevelName(self.getEffectiveLevel()).lower()


def is_enable_for(self, level):
    return _get_level_number(level) >= self.getEffectiveLevel()


logger.set_level = MethodType(set_level, logger)
logger.get_level = MethodType(get_level, logger)
logger.is_enable_for = MethodType(is_enable_for, logger)

logger.set_level("info")



import beniget
import gast as ast
import astunparse

from capturex import CaptureX

examples = {6: "issue6_import.py", 7: "issue7_func.py"}


with open(examples[6]) as file:
    code = file.read()

module = ast.parse(code)

duc = beniget.DefUseChains()
duc.visit(module)

ancestors = beniget.Ancestors()
ancestors.visit(module)

boost = [d for d in duc.locals[module] if d.name() == "boost"][0]

for user in boost.users():
    # we're interested in the parent of the decorator
    fdef = ancestors.parent(user.node)

chain = duc.chains[fdef]

for node in fdef.decorator_list:
    if node.id == "boost":
        fdef.decorator_list.remove(node)


capturex = CaptureX(module, fdef)

capturex.visit(fdef)

for node in capturex.external:
    # print(astunparse.dump(node))
    print(astunparse.unparse(node).strip())



import gast as ast
import beniget
import astunparse


class CaptureX(ast.NodeVisitor):
    def __init__(self, module_node, fun):
        self.fun = fun
        # initialize use-def chains
        self.du_chains = du = beniget.DefUseChains()
        du.visit(module_node)
        self.ud_chains = beniget.UseDefChains(du)
        self.ancestors = beniget.Ancestors()
        self.ancestors.visit(module_node)
        self.external = []
        self.visited_external = set()

    def visit_Name(self, node):
        # register load of identifiers not locally defined
        if isinstance(node.ctx, ast.Load):
            def_ = self.ud_chains.chains[node]
            try:
                parents = self.ancestors.parents(def_.node)
            except KeyError:
                return  # a builtin
            if self.fun not in parents:
                try:
                    defining_node = self.ancestors.parentStmt(def_.node)
                except ValueError:
                    # FunctionDef
                    defining_node = def_.node
                if defining_node not in self.visited_external:
                    self.rec(defining_node)
                    self.visited_external.add(defining_node)
                    self.external.append(defining_node)

    def rec(self, node):
        "walk definitions to find their operands's def"
        if isinstance(node, ast.Assign):
            self.visit(node.value)
        elif isinstance(node, ast.FunctionDef):
            old_func = self.fun
            self.fun = node
            self.visit(node)
            self.fun = old_func

        # TODO: implement this for AugAssign etc


if __name__ == "__main__":

    code = "a = 1; b = [a, a]\ndef foo():\n return b"
    # code = "a = 1; b = len([a, a])\ndef foo():\n return b"
    # code = "import numpy as np\na = np.int(1)\ndef foo():\n return np.zeros(a)"

    code = """

a = 1

def fooo():
    return 1

def foo():
    return a + fooo()

def bar():
    return foo()

    """


    module = ast.parse(code)
    function = module.body[3]
    capturex = CaptureX(module, function)

    capturex.visit(function)
    for node in capturex.external:
        # print(astunparse.dump(node))
        print(astunparse.unparse(node).strip())




from functools import wraps

import numpy as np

# pythran export __method__Transmitter____call__(float, float[])


def __for_method__Transmitter____call__(self_freq, inp):
    return inp * np.exp(np.arange(len(inp))*self_freq*1j)


def __method__Transmitter____call__(self, inp):
    self_freq = self.freq
    return __for_method__Transmitter____call__(self_freq, inp)


def my_decor(func):
    return wraps(func)(__method__Transmitter____call__)


class Transmitter():

    freq: float

    def __init__(self, freq):
        self.freq = float(freq)

    @my_decor
    def __call__(self, inp: "float[]"):
        """My docstring"""
        return inp * np.exp(np.arange(len(inp))*self.freq*1j)


"""Use Pythran to create extensions
===================================

User API
--------

.. autofunction:: wait_for_all_extensions

Internal API
------------

.. autofunction:: make_hex

.. autoclass:: SchedulerPopen
   :members:
   :private-members:

.. autofunction:: name_ext_from_path_backend

.. autofunction:: compile_extensions

.. autofunction:: compile_extension

"""

import multiprocessing
import subprocess
import time
from typing import Union, Iterable, Optional
import sysconfig
import hashlib
import sys
import os
from datetime import datetime

from .compat import open, implementation
from . import mpi
from .mpi import Path, PathSeq
from .log import logger

ext_suffix = sysconfig.get_config_var("EXT_SUFFIX") or ".so"


def modification_date(filename):
    """Get the modification date of a file"""
    return datetime.fromtimestamp(os.path.getmtime(str(filename)))


def has_to_build(output_file: Path, input_file: Path):
    """Check if a file has to be (re)built"""
    output_file = PathSeq(output_file)
    input_file = PathSeq(input_file)
    if not output_file.exists():
        return True
    mod_date_output = modification_date(output_file)
    if mod_date_output < modification_date(input_file):
        return True
    return False


def make_hex(src):
    """Produce a hash from a sting"""
    return hashlib.md5(src.encode("utf8")).hexdigest()


def name_ext_from_path_backend(path_backend):
    """Return an extension name given the path of a Pythran file"""

    name = None
    if mpi.rank == 0:
        path_backend = PathSeq(path_backend)
        if path_backend.exists():
            with open(path_backend) as file:
                src = file.read()
        else:
            src = ""

        name = path_backend.stem + "_" + make_hex(src) + ext_suffix

    return mpi.bcast(name)


class SchedulerPopen:
    """Limit the number of Pythran compilations performed in parallel

    """

    deltat = 0.2

    def __init__(self, parallel=True):
        if mpi.rank > 0:
            return
        self.processes = []
        if parallel:
            self.limit_nb_processes = max(1, multiprocessing.cpu_count() // 2)
        else:
            self.limit_nb_processes = 1

    def block_until_avail(self, parallel=True):

        if mpi.rank == 0:
            if parallel:
                limit = self.limit_nb_processes
            else:
                limit = 1

            while len(self.processes) >= limit:
                time.sleep(self.deltat)
                self.processes = [
                    process
                    for process in self.processes
                    if process.is_alive_root()
                ]

        mpi.barrier(timeout=None)

    def wait_for_all_extensions(self):
        """Wait until all compilation processes are done"""
        if mpi.rank == 0:
            while self.processes:
                time.sleep(self.deltat)
                self.processes = [
                    process
                    for process in self.processes
                    if process.is_alive_root()
                ]

        mpi.barrier(timeout=None)

    def compile_extension(
        self,
        path: Path,
        name_ext_file: Optional[str] = None,
        native=True,
        xsimd=True,
        openmp=False,
        str_pythran_flags: Optional[str] = None,
        parallel=True,
        force=True,
    ):

        if name_ext_file is None:
            name_ext_file = name_ext_from_path_backend(path)

        if not force:
            path_out = path.with_name(name_ext_file)
            if not has_to_build(path_out, path):
                logger.warning(
                    f"Do not pythranize {path} because it seems up-to-date "
                    "(but the compilation options may have changed). "
                    "You can force the compilation with the option -f."
                )
                return

        if mpi.rank == 0:
            logger.info(f"Schedule pythranization of file {path}")

        if str_pythran_flags is not None:
            flags = str_pythran_flags.strip().split()
        else:
            flags = []

        def update_flags(flag):
            if flag not in flags:
                flags.append(flag)

        if native and os.name != "nt":
            update_flags("-march=native")

        if xsimd:
            update_flags("-DUSE_XSIMD")

        if openmp:
            update_flags("-fopenmp")

        if logger.is_enable_for("debug"):
            update_flags("-v")

        words_command = [
            sys.executable,
            "-m",
            "transonic_cl.run_pythran",
            path.name,
        ]

        words_command.extend(("-o", name_ext_file))

        words_command.extend(flags)

        cwd = path.parent
        if implementation == "PyPy":
            cwd = str(cwd)
            words_command = [str(word) for word in words_command]

        self.block_until_avail(parallel)

        process = None
        if mpi.rank == 0:
            process = subprocess.Popen(words_command, cwd=cwd)

        process = mpi.ShellProcessMPI(process)

        if mpi.rank == 0:
            self.processes.append(process)
        return process


scheduler = SchedulerPopen()


def wait_for_all_extensions():
    """Wait until all compilation processes are done"""
    scheduler.wait_for_all_extensions()


def compile_extensions(
    paths: Iterable[Path], str_pythran_flags: str, parallel=True, force=True
):
    for path in paths:
        scheduler.compile_extension(
            path,
            str_pythran_flags=str_pythran_flags,
            parallel=parallel,
            force=force,
        )


def compile_extension(
    path: Union[Path, str],
    name_ext_file: Optional[str] = None,
    native=True,
    xsimd=True,
    openmp=False,
):
    if not isinstance(path, Path):
        path = Path(path)

    # return the process
    return scheduler.compile_extension(
        path, name_ext_file, native=native, xsimd=xsimd, openmp=openmp
    )



"""
See https://bitbucket.org/fluiddyn/fluidpythran/issues/6/

Here, the function uses np (numpy), but it should of course work for all
packages.

Related question: do we support usage of global variables in "boosted"
functions?

"""
import numpy as np

from transonic import boost

# this one has nothing to do with my_constant used in func
my_constant = 10

# this code produces my_constant used in func
values = [2, 3, 4]
my_constant = len(values)


@boost
def func(a: float, b: float):

    c = a + b + values[0]
    return (a * np.log(c)).max() * my_constant


if __name__ == "__main__":
    result1 = func(1.0, 1.0)
    result2 = func(2.0, 2.0)



from pathlib import Path

# import ast
import gast as ast
import beniget
import astunparse
from pprint import pprint

from capturex import CaptureX, make_code_external


def dump(node):
    print(astunparse.dump(node))


path_examples = Path("examples")

files = sorted(path_examples.glob("*.py"))

with open(files[3]) as file:
    code = file.read()

module = ast.parse(code)

duc = beniget.DefUseChains()
duc.visit(module)

ancestors = beniget.Ancestors()
ancestors.visit(module)

udc = beniget.UseDefChains(duc)


functions_boosted = {}
class_boosted = {}
methods_boosted = {}

# we first need to find the node where transonic.boost is defined...


def add_definition(definition_node):
    tmp_boosted = None
    key = definition_node.name
    if isinstance(definition_node, ast.FunctionDef):
        parent = ancestors.parent(definition_node)
        if isinstance(parent, ast.ClassDef):
            tmp_boosted = methods_boosted
            key = (parent.name, key)
        else:
            tmp_boosted = functions_boosted
    elif isinstance(definition_node, ast.ClassDef):
        tmp_boosted = class_boosted
    if tmp_boosted is not None:
        tmp_boosted[key] = definition_node


for node in module.body:
    if isinstance(node, ast.Import):
        for alias in node.names:
            if alias.name == "transonic":
                transonic_def_node = alias
                transonic_def = duc.chains[transonic_def_node]
                for user in transonic_def.users():
                    for user1 in user.users():
                        if (
                            isinstance(user1.node, ast.Attribute)
                            and user1.node.attr == "boost"
                        ):
                            definition_node = ancestors.parent(user1.node)
                            add_definition(definition_node)
    elif isinstance(node, ast.ImportFrom):
        if node.module == "transonic":
            for alias in node.names:
                if alias.name == "boost":
                    boost_def_node = alias
                    boost_def = duc.chains[boost_def_node]
                    for user in boost_def.users():
                        definition_node = ancestors.parent(user.node)
                        add_definition(definition_node)


pprint(functions_boosted)
pprint(methods_boosted)
pprint(class_boosted)

objects = []

for boosted in (functions_boosted, methods_boosted):
    for fdef in boosted.values():
        fdef.decorator_list = []
        objects.append(fdef)

for cdef in class_boosted.values():
    objects.append(cdef)

capturex_annotations = CaptureX(
    objects,
    module,
    defuse_chains=duc,
    usedef_chains=udc,
    ancestors=ancestors,
    consider_annotations="only",
)

code_dependance_annotations = make_code_external(capturex_annotations.external)

capturex = CaptureX(
    objects,
    module,
    defuse_chains=duc,
    usedef_chains=udc,
    ancestors=ancestors,
    consider_annotations=False,
)

code_dependance = make_code_external(capturex.external)



import numpy as np

from transonic import boost

type_ = float

@boost
class Transmitter:

    freq: type_

    def __init__(self, freq):
        self.freq = float(freq)

    @boost
    def __call__(self, inp: "float[]"):
        """My docstring"""
        return inp * np.exp(np.arange(len(inp)) * self.freq * 1j)


if __name__ == "__main__":
    inp = np.ones(2)
    freq = 1.0
    trans = Transmitter(freq)

    def for_check(freq, inp):
        return inp * np.exp(np.arange(len(inp)) * freq * 1j)

    assert np.allclose(trans(inp), for_check(freq, inp))
